//***************************************************************************
// Copyright 2007-2016 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// http://ec.europa.eu/idabc/eupl.html.                                     *
//***************************************************************************
// Author: Andreas Vibeto                                                   *
//***************************************************************************

#include <cmath>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
//#include "path.hpp"

// DUNE headers.
#include <DUNE/DUNE.hpp>

namespace FixedCam
{
  namespace Path
  {
    using DUNE_NAMESPACES;
    using namespace std;

    enum States
    {
      IDLE = -1,
      GO_TO_START_POINT,
      GO_TO_WPP,
      CALCULATE_NEXT_WPP
    };

    struct Task: public DUNE::Tasks::Task
    {
      //! Constructor.
      //! @param[in] name task name.
      //! @param[in] ctx context.


      // State for state-machine
      States m_current_state;

      bool m_is_on_ground;
      bool m_is_executing;
      bool m_is_ready;
      bool m_is_initiated;

      // START POSITION
      fp64_t start_x;
      fp64_t start_y;
      fp64_t velocity, altitude;
      fp64_t lat_init, lon_init;

      // Path Start
      fp64_t lat_path_start;
      fp64_t lon_path_start;

      int lookahead_dist;

      // Path data
      string filename;
      fp64_t path_length;
      double **path_data;
      
      int prev_i;

      // IMC messages
      IMC::Reference                m_ref;
      IMC::EstimatedState           m_estimated_state;
      //IMC::EstimatedStreamVelocity  m_ewind;
      IMC::DesiredZ                 m_dheight;
      IMC::DesiredSpeed             m_dspeed;


      Task(const std::string& name, Tasks::Context& ctx):
        DUNE::Tasks::Task(name, ctx),
        m_current_state(IDLE),
        m_is_on_ground(true),
        m_is_executing(false),
        m_is_ready(false),
        m_is_initiated(false),
        path_length(550),
        prev_i(0)
      {

        // Load parameters from .ini-file
        param("Start X", start_x)
        .defaultValue("0")
        .units(Units::Meter)
        .description("The x-coordinate for that start point");

        param("Start Y", start_y)
        .defaultValue("0")
        .units(Units::Meter)
        .description("The y-coordinate for that start point");

        param("Height", altitude)
        .defaultValue("100")
        .units(Units::Meter)
        .description("The height to use throughout the flight");

        param("Speed", velocity)
        .defaultValue("20")
        .description("The speed to use throughout the flight");

        param("Path Filename", filename)
        .defaultValue("path.txt")
        .description("Path to file containing path");
  
        //param("Path Length", path_length)
        //.defaultValue("-1")
        //.description("Number of entries in the path-file");

        param("Lookahead Distance", lookahead_dist)
        .defaultValue("50")
        .description("Lookahead distance for the path follower");

        // Read path
        ifstream file("/home/andreas/pos_lin_easy.txt");
        path_data = readPathFile(file, 550);

        //bind<EstimatedStreamVelocity>(this); // Wind States
        bind<EstimatedState>(this);          // States
        bind<VehicleMedium>(this);           // Is vehicle on ground?
        bind<PlanControlState>(this);        // Something

      }

      //! Update internal state with new parameter values.
      void
      onUpdateParameters(void)
      {
      }

      //! Reserve entity identifiers.
      void
      onEntityReservation(void)
      {
      }

      //! Resolve entity names.
      void
      onEntityResolution(void)
      {
      }

      //! Acquire resources.
      void
      onResourceAcquisition(void)
      {
      }

      //! Initialize resources.
      void
      onResourceInitialization(void)
      {
      }

      //! Release resources.
      void
      onResourceRelease(void)
      {
      }

      //Check control state
      void
      consume(const IMC::PlanControlState* msg)
      {
        if(msg->state == PlanControlState::PCS_READY )
        {
          m_is_ready = true;
        }
      }

      // Get estimated state
      void
      consume(const IMC::EstimatedState* msg)
      {
        m_estimated_state = *msg;
        
        if(m_is_ready && !m_is_on_ground && !m_is_initiated){
          m_is_initiated = true;
          war("Ready for target coordinates.");
          goToStart();
          m_current_state = GO_TO_START_POINT;
        }

        stateMachine();
      }


      // Get estimated wind
      /*void
      consume(const IMC::EstimatedStreamVelocity* msg)
      {

      }*/

      // Check if in the air
      void
      consume(const IMC::VehicleMedium* vm)
      {
        m_is_on_ground = (vm->medium == IMC::VehicleMedium::VM_GROUND);
        if(m_is_on_ground){
          //war("m_ground == true!");
        }else{
           if(!m_is_executing){
              startExecution();
           }
        }
      }

      // Start execution: Magic that causes stuff to work
      void
      startExecution( void )
      {
        m_is_executing = true;
        IMC::PlanControl startPlan;
        startPlan.type = IMC::PlanControl::PC_REQUEST;
        startPlan.op = IMC::PlanControl::PC_START;
        startPlan.plan_id = "fixedCam-path";

        IMC::FollowReference man;
        man.control_ent = getEntityId();
        man.control_src = getSystemId();
        man.altitude_interval = 5;
        man.timeout = 0.0;

        IMC::PlanSpecification spec;
        spec.plan_id = "fixedCam-path";
        spec.start_man_id = "follow-path";
        
        IMC::PlanManeuver pm;
        pm.data.set(man);
        pm.maneuver_id = "follow-path";
        pm.start_actions = maneuverParameters(false);
        spec.maneuvers.push_back(pm);
        startPlan.arg.set(spec);
        startPlan.request_id = 0;
        startPlan.flags = 0;

        dispatch( startPlan );
      }
    
      //Set controller (path or waypoint): Magic that causes stuff to work
      IMC::MessageList<IMC::Message>
      maneuverParameters(bool bank)
      {
        IMC::MessageList<IMC::Message> setEntityParameters;

        IMC::SetEntityParameters* sep = new IMC::SetEntityParameters();
        sep->name = "Path Control";
        IMC::MessageList<IMC::EntityParameter> entityParameters;
        IMC::EntityParameter* ep = new IMC::EntityParameter();
        ep->name = "Use controller";
        ep->value = bank ? "true" : "false";
        entityParameters.push_back(*ep);
        delete ep;
        sep->params = entityParameters;
        setEntityParameters.push_back(*sep);
        delete sep;

        sep = new IMC::SetEntityParameters();
        sep->name = "Height Control";
        entityParameters.clear();
        ep = new IMC::EntityParameter();
        ep->name = "Active";
        ep->value = bank ? "true" : "false";
        entityParameters.push_back(*ep);
        delete ep;
        sep->params = entityParameters;
        setEntityParameters.push_back(*sep);
        delete sep;

        sep = new IMC::SetEntityParameters();
        sep->name = "Autopilot";
        entityParameters.clear();
        ep = new IMC::EntityParameter();
        ep->name = "Ardupilot Tracker";
        ep->value = bank ? "false" : "true";
        entityParameters.push_back(*ep);
        delete ep;

        ep = new IMC::EntityParameter();
        ep->name = "Formation Flight";
        ep->value = "false";
        entityParameters.push_back(*ep);
        delete ep;
        sep->params = entityParameters;
        setEntityParameters.push_back(*sep);
        delete sep;

        return setEntityParameters;
      }

      // Go to start point
      void
      goToStart( void )
      {
        war("Go to start");
        //fp64_t lat, lon;
        getCurrentLatLon(&lat_init, &lon_init);
        WGS84::displace(start_x, start_y, &lat_init, &lon_init);

        setNewPosRef(lat_init, lon_init);
      }


      // Calculate if at start
      void
      distanceToStart( void )
      {
        fp64_t dist_x = start_x - m_estimated_state.x;
        fp64_t dist_y = start_y - m_estimated_state.y;
        fp64_t dist   = sqrt(dist_x*dist_x + dist_y*dist_y);

        fp64_t height_diff = abs(altitude - getCurrentHeight());

        //inf("Distance to start point: %f, height diff: %f", dist, height_diff);
        
        setNewPosRef(lat_init, lon_init);        

        if((dist < 200) && (height_diff < 50)){
          war("Close to start point!");

          // Calculate path start point
          gotoPathStart();

          m_current_state = GO_TO_WPP;
        }
      }

      // Set new reference with lat and lon
      void
      setNewPosRef(fp64_t lat, fp64_t lon)
      {
        // Caluclate displaced coordinates
        //fp64_t lat, lon;
        //getCurrentLatLon(&lat, &lon);
        //WGS84::displace(x, y, &lat, &lon);

        // Set reference message
        m_dspeed.value  = velocity;
        m_dheight.value = altitude;

        m_ref.radius = 10;
        m_ref.lat    = lat;
        m_ref.lon    = lon;
        m_ref.z.set(m_dheight);
        m_ref.speed.set(m_dspeed);
        m_ref.flags  = Reference::FLAG_DIRECT   |
                       Reference::FLAG_LOCATION |
                       Reference::FLAG_RADIUS   |
                       Reference::FLAG_SPEED    |
                       Reference::FLAG_Z        ;

        //inf("Setting new pos ref: %f, %f", lat, lon);

        if(!m_is_on_ground){
          dispatch(m_ref);
        }
      }

      // Go to path start
      void
      gotoPathStart()
      {
        war("go to path start");
        //fp64_t lat, lon;
        getCurrentLatLon(&lat_path_start, &lon_path_start);
        WGS84::displace(500, 0, &lat_path_start, &lon_path_start);
      
        //lat_path_start = lat;
        //lon_path_start = lon;
        err("Path start: %f , %f", lat_path_start, lon_path_start);
        setNewPosRef(lat_path_start, lon_path_start);

      }

      // Get current lat and lon
      void
      getCurrentLatLon(fp64_t *lat, fp64_t *lon)
      {
        *lat = m_estimated_state.lat;
        *lon = m_estimated_state.lon;

        WGS84::displace(m_estimated_state.x, m_estimated_state.y,
                        lat, lon);
      }

      // Get current height
      fp32_t
      getCurrentHeight()
      {
        return -m_estimated_state.z+m_estimated_state.height;
      }

      // State Machine
      void
      stateMachine( void )
      {
        switch( m_current_state )
        {
          case IDLE:
            break;
        
          case GO_TO_START_POINT:
            //war("GO TO START");
            //goToStart();
            distanceToStart();
            break;

          case GO_TO_WPP:
            //war("GO TO WPP");
            findNewWpp();
            break;

          case CALCULATE_NEXT_WPP:
            break;

          default:
            return;
        }
      }

      // Find the next wpp
      void
      findNewWpp()
      {
        //err("Find new wpp");
        int point_found = 0;

        //for(int i = path_length-1; i >= 0 ; i--){
        for(int i = prev_i; i < path_length ; i++){
          float dist = distanceFromCurrentPos(path_data[i][0], path_data[i][1]);
          inf("Dist: %f, i= %i", dist, i);
          if(dist > lookahead_dist){

            fp64_t lat = lat_path_start;
            fp64_t lon = lon_path_start;
            WGS84::displace(path_data[i][0], path_data[i][1], &lat, &lon);

            //war("Found point! %f , %f  Dist: %f", lat, lon, dist);
            war("NED: %f , %f  : (%i) : %f", path_data[i][0], path_data[i][1], i, dist);

            setNewPosRef(lat, lon);
            point_found = 1;
            prev_i = i;
            break;
          }
        }
        if(!point_found){
          err("Could not find point!");
        }
      }

      // Calculate distance from current pos to point on path
      int
      distanceFromCurrentPos(fp64_t x, fp64_t y)
      {
        // Calculate path lat and lon
        fp64_t path_lat = lat_path_start;
        fp64_t path_lon = lon_path_start;
        WGS84::displace(x, y, &path_lat, &path_lon);

        // Get current lat and lon
        fp64_t cur_lat, cur_lon;
        getCurrentLatLon(&cur_lat, &cur_lon);

        // Get distance and bearing between points
        fp64_t bearing, distance;
        WGS84::getNEBearingAndRange(path_lat, path_lon, cur_lat, cur_lon,
                                                      &bearing, &distance);

        return distance;
      }

      double**
      readPathFile(ifstream& file, int path_len){
        war("Reading path file");
        /* Read file to array */
        double** path;
        path = new double*[path_len];
        for( int i = 0 ; i < path_len ; i++ ){
          path[i] = new double[2];
        }
        for( int row = 0 ; row < path_len ; ++row ){
          string line;
          getline(file, line);
        
          if( !file.good() ){
            err("readPathFile: Something went wrong when reading" \
                        " a line in the path file! %i \n", row);            
            break;
          }

          stringstream iss(line);

          string x_value;
 
          getline(iss, x_value, ',');
          if( !iss.good() ){
            err("readPathFile: Something went wrong when reading" \
                        " the x-value in the path file!\n");
            break;
          }
        
          stringstream x_convertor(x_value);
          x_convertor >> path[row][0];

          string y_value;
          getline(iss, y_value, '\n');

          stringstream y_convertor(y_value);
          y_convertor >> path[row][1];
        }
    
        return path;
      }

      //! Main loop.
      void
      onMain(void)
      {
        while (!stopping())
        {
          waitForMessages(1.0);
        }
      }
    };
  }
}

DUNE_TASK
